/* Greedy adversary
 * For each call to wait() we return the outcome that yields the maximal number
 * of possible locations compatible with our replies.
 * Note that in the special case H=1 this means that we will find a collision if
 * it exists, so we can grade the 'encode part' of the user program perfectly!
 *
 * Running time: O(N²H)
 * Memory:       O(N²)
 */

#pragma once

#include"offline.h"
#include"../manager_constants.h"
#include<vector>
#include<algorithm>
#include<unordered_map>
#include<random>
#include<cassert>

using namespace std;

class greedy_adversary : public offline_adversary
{
 public:
    greedy_adversary(int s) : outcomes(N + 1, 0), actual_outcome(0),
                              mask(1), reply_mask(1)
    {
        rnd_engine.seed(s);
    }

    bool answer(int x, int y)
    {
        if(x < y) swap(x, y);
    
        for(int a = 1; a <= N; ++a)
        {
            for(int b = 1; b <=a; ++b)
            {
                bitstring curr_outcome = outcomes[a] | outcomes[b];

                if(curr_outcome == actual_outcome and (a != x or b != y))
                {
                    #ifdef DEBUG
                        fprintf(stderr, "[selecting %d,%d instead of %d,%d]", 
                                a, b, x, y);
                    #endif

                    return false;
                }
            }
        }
        
        return true;
    }

    typedef __int128 bitstring;
    
 protected:
    vector<int> get_replies(vector<vector<int>> queries)
    {
        bitstring cutoff = mask - 1;
        
        update_outcomes(queries);

        unordered_map<bitstring, int> block_size;
        int max_block_size = 0;

        for(int x = 1; x <= N; ++x)
        {
            for(int y = 1; y <= x; ++y)
            {
                bitstring curr_outcome = outcomes[x] | outcomes[y];
                
                if((curr_outcome & cutoff) != actual_outcome)
                    continue;
            
                ++block_size[curr_outcome];
                max_block_size = max(max_block_size, block_size[curr_outcome]);
            }
        }

        vector<bitstring> possible_outcomes;
        for(auto b : block_size)
            if(b.second == max_block_size)
                possible_outcomes.push_back(b.first);
        assert(possible_outcomes.size() > 0);

        uniform_int_distribution<int> distr(0, possible_outcomes.size() - 1);
        actual_outcome = possible_outcomes[distr(rnd_engine)];

        #ifdef DEBUG
            fprintf(stderr, "[block size %d]", block_size[actual_outcome]);
        #endif

        return reply_from_bitstring(actual_outcome, queries.size());
    }

 private:
    vector<bitstring> outcomes;
    bitstring actual_outcome;
    
    bitstring mask;
    bitstring reply_mask;    
    
    mt19937 rnd_engine;

    // The following two functions must each be called exactly once in in
    // get_replies and in this order!
    void update_outcomes(vector<vector<int>> queries)
    {
        for(int i = 0; i < (int) queries.size(); ++i)
        {
            for(int x : queries[i])
                outcomes[x] += mask;
            mask <<= 1;
        }
    }
        
    vector<int> reply_from_bitstring(bitstring s, int length)
    {
        vector<int> reply;
    
        for(int i = 0; i < length; ++i)
        {
            reply.push_back(s & reply_mask ? 1 : 0);
            reply_mask <<= 1;
        }
        
        return reply;
    }

};
