/* Adversary for the case of a single scout (and hence H=1)
 * We will find a collision if it exists, so we can grade the 'encode part' of
 * the user program perfectly!
 *
 * Running time: O(NR)
 * Memory:       O(N)
 */

#pragma once

#include"offline.h"
#include"../manager_constants.h"
#include<vector>
#include<algorithm>
#include<unordered_set>
#include<random>
#include<cassert>

using namespace std;
typedef __int128 bitstring;

class single_adversary : public offline_adversary
{
 public:
    single_adversary(int s) : outcomes(N + 1, 0)
    {
        rnd_engine.seed(s);
    }

    bool answer(int x, int y)
    {
        if(x != y) return false;
    
        for(int a = 1; a <= N; ++a)
        {
            bitstring curr_outcome = outcomes[a];

            if(curr_outcome == actual_outcome and a != x)
            {
                #ifdef DEBUG
                    fprintf(stderr, "[selecting %d instead of %d]", a, x);
                #endif

                return false;
            }
        }
        
        return true;
    }
    
 protected:
    vector<int> get_replies(vector<vector<int>> queries)
    {
        bitstring mask = 1;
        
        for(int i = 0; i < (int) queries.size(); ++i)
        {
            for(int x : queries[i])
                outcomes[x] += mask;
            mask <<= 1;
        }

        unordered_set<bitstring> results;

        for(int x = 1; x <= N; ++x)
        {
            bitstring curr_outcome = outcomes[x];
            
            if(results.find(curr_outcome) == results.end())
            {
                results.insert(curr_outcome);
            }
            
            else
            {
                // We could just decide to not accept the answer at this
                // point -- however, we will instead actually find a
                // different possible set of locations in answer (this might
                // be helpful for debugging purposes)
                actual_outcome = curr_outcome;
                return reply_from_bitstring(actual_outcome, queries.size());
            }
        }

        uniform_int_distribution<int> dis(1, 1000);        
        actual_outcome = outcomes[dis(rnd_engine)];
        return reply_from_bitstring(actual_outcome, queries.size());
    }

 private:
    vector<bitstring> outcomes;
    bitstring actual_outcome;

    mt19937 rnd_engine;

    vector<int> reply_from_bitstring(bitstring s, int length)
    {
        vector<int> reply;
        bitstring mask = 1;
    
        for(int i = 0; i < length; ++i)
        {
            reply.push_back(s & mask ? 1 : 0);
            mask <<= 1;
        }
        
        return reply;
    }
};
