/* An intermediate class for H=2 strategies
 * The strategy for the first call to wait() is to be implemented by the
 * subclass. For the second call to wait() we look for a collision: if one
 * exists, we take this one (killing the submission), otherwise we take some
 * random consistent outcome
 *
 * Running time: O(N²) + time for subclass
 * Memory:       O(N²) + memory for subclass
 */

#pragma once

#include"offline.h"
#include"../manager_constants.h"
#include<vector>
#include<algorithm>
#include<unordered_map>
#include<random>
#include<cassert>

using namespace std;

class two_phase_adversary : public offline_adversary
{
 public:
    two_phase_adversary(int s) : outcomes(N + 1, 0), actual_outcome(0),
                                 mask(1), reply_mask(1), first_run(true)
    {
        rnd_engine.seed(s);
    }

    bool answer(int x, int y)
    {
        if(x < y) swap(x, y);
    
        for(int a = 1; a <= N; ++a)
        {
            for(int b = 1; b <=a; ++b)
            {
                bitstring curr_outcome = outcomes[a] | outcomes[b];

                if(curr_outcome == actual_outcome and (a != x or b != y))
                {
                    #ifdef DEBUG
                        fprintf(stderr, "[selecting %d,%d instead of %d,%d]", 
                                a, b, x, y);
                    #endif

                    return false;
                }
            }
        }
        
        return true;
    }

    typedef __int128 bitstring;
    typedef unordered_map<bitstring, int> block_map;
    
 protected:
    mt19937 rnd_engine;
    vector<bitstring> outcomes;

    vector<int> get_replies(vector<vector<int>> queries)
    {
        bitstring cutoff = mask - 1;
        
        update_outcomes(queries);

        block_map block_size;

        for(int x = 1; x <= N; ++x)
        {
            for(int y = 1; y <= x; ++y)
            {
                bitstring curr_outcome = outcomes[x] | outcomes[y];
                
                if((curr_outcome & cutoff) != actual_outcome)
                    continue;
            
                ++block_size[curr_outcome];
            }
        }

        if(first_run) actual_outcome = first_reply(block_size, queries);
        else          actual_outcome = second_reply(block_size, cutoff);

        first_run = false;

        return reply_from_bitstring(actual_outcome, queries.size());
    }

    virtual bitstring first_reply(const block_map&, vector<vector<int>>) = 0; // to be defined in subclass
    
    virtual bitstring second_reply(const block_map& block_size, bitstring cutoff)
    {
        vector<bitstring> possible_outcomes;

        for(auto b : block_size)
        {
            if((b.first & cutoff) != actual_outcome)
                continue;

            possible_outcomes.push_back(b.first);
        
            if(b.second > 1)
            {
                #ifdef DEBUG
                    fprintf(stderr, "[block size %d]", b.second);
                #endif
                
                return b.first;
            }
        }       

        assert(not possible_outcomes.empty());

        uniform_int_distribution<int> distr(0, possible_outcomes.size() - 1);
        return possible_outcomes[distr(rnd_engine)];
    }

 private:
    bitstring actual_outcome;
    
    bitstring mask;
    bitstring reply_mask;
    
    bool first_run;  

    // The following two functions must each be called exactly once in in
    // get_replies and in this order!
    void update_outcomes(vector<vector<int>> queries)
    {
        for(int i = 0; i < (int) queries.size(); ++i)
        {
            for(int x : queries[i])
                outcomes[x] += mask;
            mask <<= 1;
        }
    }
        
    vector<int> reply_from_bitstring(bitstring s, int length)
    {
        vector<int> reply;
    
        for(int i = 0; i < length; ++i)
        {
            reply.push_back(s & reply_mask ? 1 : 0);
            reply_mask <<= 1;
        }
        
        return reply;
    }

};
